!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #######################
     MODULE MODI_TURB_HOR_VW
!    #######################
!
INTERFACE  
!
      SUBROUTINE TURB_HOR_VW(KSPLT,                                  &
                      OTURB_FLX,KRR,                                 &
                      TPFILE,                                        &
                      PK,PINV_PDYY,PINV_PDZZ,PMZM_PRHODJ,            &
                      PDYY,PDZZ,PDZY,                                &
                      PRHODJ,PTHVREF,                                &
                      PVM,PWM,PTHLM,PRM,PSVM,                        &
                      PTKEM,PLM,                                     &
                      PDP,                                           &
                      PRVS,PRWS                                      )
!
USE MODD_IO, ONLY: TFILEDATA
!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                  INTENT(IN)    ::  KRR          ! number of moist var.
TYPE(TFILEDATA),          INTENT(INOUT) ::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDYY   ! 1./PDYY
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDYY, PDZZ, PDZY 
                                                         ! Metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTHVREF      ! ref. state VPT       
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PVM,PWM,PTHLM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PSVM
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTKEM        ! TKE at time t- dt
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PLM          ! Turb. mixing length
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRVS, PRWS   ! var. at t+1 -split-
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!
END SUBROUTINE TURB_HOR_VW
!
END INTERFACE
!
END MODULE MODI_TURB_HOR_VW
!     ################################################################
      SUBROUTINE TURB_HOR_VW(KSPLT,                                  &
                      OTURB_FLX,KRR,                                 &
                      TPFILE,                                        &
                      PK,PINV_PDYY,PINV_PDZZ,PMZM_PRHODJ,            &
                      PDYY,PDZZ,PDZY,                                &
                      PRHODJ,PTHVREF,                                &
                      PVM,PWM,PTHLM,PRM,PSVM,                        &
                      PTKEM,PLM,                                     &
                      PDP,                                           &
                      PRVS,PRWS                                      )
!     ################################################################
!
!
!!****  *TURB_HOR* -routine to compute the source terms in the meso-NH
!!               model equations due to the non-vertical turbulent fluxes.
!!
!!    PURPOSE
!!    -------
!!
!!     see TURB_HOR
!!
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!                     Aug    , 1997 (V. Saravane) spliting of TURB_HOR
!!                     Nov  27, 1997 (V. Masson) clearing of the routine
!!                     Oct  18, 2000 (V. Masson) LES computations + LFLAT switch
!!                     Feb  14, 2001 (V. Masson and J. Stein) DZF bug on PRWS
!!                                + remove the use of W=0 at the ground
!!                                + extrapolataion under the ground
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     October 2009 (G. Tanguy) add ILENCH=LEN(YCOMMENT) after
!!                                              change of YCOMMENT
!!                     04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!!  Philippe Wautelet: 05/2016-04/2018: new data structures and calls for I/O
!! --------------------------------------------------------------------------
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_CONF
USE MODD_CTURB
use modd_field,          only: tfielddata, TYPEREAL
USE MODD_IO,             ONLY: TFILEDATA
USE MODD_PARAMETERS
USE MODD_LES
USE MODD_NSV
!
USE MODE_IO_FIELD_WRITE, only: IO_Field_write
use mode_mppdb
!
USE MODI_GRADIENT_M
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_COEFJ
USE MODI_LES_MEAN_SUBGRID
!
USE MODI_SECOND_MNH

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK , ONLY : MNH_ALLOCATE_ZT3D , MNH_REL_ZT3D, MNH_ALLOCATE_ZT3DP , MNH_ALLOCATE_ZT2D
#endif

IMPLICIT NONE
!
!
!*       0.1  declaration of arguments
!
!
!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                  INTENT(IN)    ::  KRR          ! number of moist var.
TYPE(TFILEDATA),          INTENT(INOUT) ::  TPFILE       ! Output file
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDYY   ! 1./PDYY
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDYY, PDZZ, PDZY 
                                                         ! Metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTHVREF      ! ref. state VPT       
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PVM,PWM,PTHLM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PSVM
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTKEM        ! TKE at time t- dt
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PLM          ! Turb. mixing length
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRVS, PRWS   ! var. at t+1 -split-
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!
!*       0.2  declaration of local variables
!
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZFLX,ZWORK ! work arrays
!! REAL, DIMENSION(:,:,:), pointer , contiguous :: ZVPTV
INTEGER             :: IKB,IKE,IKU
                                    ! Index values for the Beginning and End
                                    ! mass points of the domain  
INTEGER             :: JSV          ! scalar loop counter
!
REAL, DIMENSION(:,:,:), pointer , contiguous :: GY_W_VW_PWM
#ifdef MNH_OPENACC
INTEGER :: IZFLX,IZWORK,IGY_W_VW_PWM
#endif
!
REAL :: ZTIME1, ZTIME2
!
#ifdef MNH_OPENACC
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP1_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP2_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP3_DEVICE
REAL, DIMENSION(:,:,:), pointer , contiguous :: ZTMP4_DEVICE
INTEGER :: IZTMP1_DEVICE, IZTMP2_DEVICE, IZTMP3_DEVICE, IZTMP4_DEVICE
#endif
TYPE(TFIELDDATA) :: TZFIELD
!
INTEGER  :: JIU,JJU,JKU
INTEGER  :: JI,JJ,JK
! ---------------------------------------------------------------------------

!$acc data present( PK, PINV_PDYY, PINV_PDZZ, PMZM_PRHODJ, PDYY, PDZZ, PDZY, &
!$acc &             PRHODJ, PTHVREF,                                         &
!$acc &             PVM, PWM, PTHLM, PRM, PSVM, PTKEM, PLM,                  &
!$acc &             PRVS, PRWS, PDP)

if ( mppdb_initialized ) then
  !Check all in arrays
  call Mppdb_check( pk,          "Turb_hor_vw beg:pk"          )
  call Mppdb_check( pinv_pdyy,   "Turb_hor_vw beg:pinv_pdyy"   )
  call Mppdb_check( pinv_pdzz,   "Turb_hor_vw beg:pinv_pdzz"   )
  call Mppdb_check( pdyy,        "Turb_hor_vw beg:pdyy"        )
  call Mppdb_check( pdzz,        "Turb_hor_vw beg:pdzz"        )
  call Mppdb_check( pdzy,        "Turb_hor_vw beg:pdzy"        )
  call Mppdb_check( prhodj,      "Turb_hor_vw beg:prhodj"      )
  call Mppdb_check( pthvref,     "Turb_hor_vw beg:pthvref"     )
  call Mppdb_check( pvm,         "Turb_hor_vw beg:pvm"         )
  call Mppdb_check( pwm,         "Turb_hor_vw beg:pwm"         )
  call Mppdb_check( pthlm,       "Turb_hor_vw beg:pthlm"       )
  call Mppdb_check( prm,         "Turb_hor_vw beg:prm"         )
  call Mppdb_check( psvm,        "Turb_hor_vw beg:psvm"        )
  call Mppdb_check( ptkem,       "Turb_hor_vw beg:ptkem"       )
  call Mppdb_check( plm,         "Turb_hor_vw beg:plm"         )
  !Check all inout arrays
  call Mppdb_check( prvs, "Turb_hor_vw beg:prvs" )
  call Mppdb_check( prws, "Turb_hor_vw beg:prws" )
  call Mppdb_check( pdp,  "Turb_hor_vw beg:pdp"  )
end if

JIU =  size(pwm, 1 )
JJU =  size(pwm, 2 )
JKU =  size(pwm, 3 )

#ifndef MNH_OPENACC
allocate( zflx (JIU,JJU,JKU ) )
allocate( zwork(JIU,JJU,JKU ) )

allocate( gy_w_vw_pwm(JIU,JJU,JKU ) )
#else
izflx        = MNH_ALLOCATE_ZT3D( zflx ,JIU,JJU,JKU)
izwork       = MNH_ALLOCATE_ZT3D( zwork,JIU,JJU,JKU)

igy_w_vw_pwm = MNH_ALLOCATE_ZT3D( gy_w_vw_pwm,JIU,JJU,JKU)
#endif

#ifdef MNH_OPENACC
iztmp1_device = MNH_ALLOCATE_ZT3D( ztmp1_device,JIU,JJU,JKU)
iztmp2_device = MNH_ALLOCATE_ZT3D( ztmp2_device,JIU,JJU,JKU)
iztmp3_device = MNH_ALLOCATE_ZT3D( ztmp3_device,JIU,JJU,JKU)
iztmp4_device = MNH_ALLOCATE_ZT3D( ztmp4_device,JIU,JJU,JKU)
#endif

!$acc data present( ZFLX, ZWORK, GY_W_VW_PWM, &
!$acc &            ZTMP1_DEVICE, ZTMP2_DEVICE, ZTMP3_DEVICE, ZTMP4_DEVICE )

!
!*       1.   PRELIMINARY COMPUTATIONS
!             ------------------------
!
IKB = 1+JPVEXT               
IKE = SIZE(PWM,3)-JPVEXT    
IKU = SIZE(PWM,3)
!
!
IF (.NOT. L2D) THEN
#ifndef MNH_OPENACC
  GY_W_VW_PWM = GY_W_VW(PWM,PDYY,PDZZ,PDZY)
#else
  CALL GY_W_VW_DEVICE(1,IKU,1,PWM,PDYY,PDZZ,PDZY,GY_W_VW_PWM)
#endif
END IF
!
!
!*      14.   < V'W'>
!             -------
!
! residual part of < V'W'> depending on dw/dy
!
IF (.NOT. L2D) THEN
#ifndef MNH_OPENACC
  ZFLX(:,:,:) =                                                      &
    - XCMFS * MYM(MZM(PK)) * GY_W_VW_PWM
#else
  CALL MZM_DEVICE(PK,ZTMP1_DEVICE)
  CALL MYM_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
  !$acc kernels
  !$acc loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZFLX(JI,JJ,JK) = - XCMFS * ZTMP2_DEVICE(JI,JJ,JK) * GY_W_VW_PWM(JI,JJ,JK)
  END DO !CONCURRENT
  !$acc end kernels
#endif
ELSE
!$acc kernels
  ZFLX(:,:,:) = 0.
!$acc end kernels
END IF
!
!$acc kernels
ZFLX(:,:,IKE+1) = 0.  ! rigid wall condition => no turbulent flux
!
!
! Nullify the flux at the ground level because it has been fully taken into
! account in turb_ver and extrapolate the flux under the ground 
ZFLX(:,:,IKB) = 0.
ZFLX(:,:,IKB-1)= 2.*ZFLX(:,:,IKB) - ZFLX(:,:,IKB+1)
!$acc end kernels
!
! stores  <V W>
IF ( tpfile%lopened .AND. OTURB_FLX ) THEN
  TZFIELD%CMNHNAME   = 'VW_HFLX'
  TZFIELD%CSTDNAME   = ''
  TZFIELD%CLONGNAME  = 'VW_HFLX'
  TZFIELD%CUNITS     = 'm2 s-2'
  TZFIELD%CDIR       = 'XY'
  TZFIELD%CCOMMENT   = 'X_Y_Z_VW_HFLX'
  TZFIELD%NGRID      = 7
  TZFIELD%NTYPE      = TYPEREAL
  TZFIELD%NDIMS      = 3
  TZFIELD%LTIMEDEP   = .TRUE.
!$acc update self(ZFLX)
  CALL IO_Field_write(TPFILE,TZFIELD,ZFLX)
END IF
!
! compute the source for rho*V due to this residual flux ( the other part is
! taken into account in TURB_VER)
#ifndef MNH_OPENACC
IF (.NOT. L2D) &
PRVS(:,:,:) = PRVS(:,:,:) - DZF( ZFLX* MYM( PMZM_PRHODJ ) / MYM ( PDZZ ) )
#else
IF (.NOT. L2D) THEN
  CALL MYM_DEVICE( PMZM_PRHODJ, ZTMP1_DEVICE )
  CALL MYM_DEVICE( PDZZ, ZTMP2_DEVICE )
  !$acc kernels
  !$acc loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     ZTMP3_DEVICE(JI,JJ,JK) = ZFLX(JI,JJ,JK)* ZTMP1_DEVICE(JI,JJ,JK) / ZTMP2_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
  CALL DZF_DEVICE( ZTMP3_DEVICE, ZTMP1_DEVICE )
  !$acc kernels
  !$acc loop independent collapse(3)
  DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
     PRVS(JI,JJ,JK) = PRVS(JI,JJ,JK) - ZTMP1_DEVICE(JI,JJ,JK)
  END DO !CONCURRENT   
  !$acc end kernels
ENDIF
#endif
!
!computation of the source for rho*W due to this flux
#ifndef MNH_OPENACC
IF (.NOT. L2D) THEN 
  IF (.NOT. LFLAT) THEN
    PRWS(:,:,:) = PRWS(:,:,:)                              &
          -DYF( MZM( MYM(PRHODJ) * PINV_PDYY) * ZFLX)           &
          +DZM( PRHODJ * MYF( MZF( ZFLX*PDZY ) * PINV_PDYY ) / MZF(PDZZ) )
  ELSE
    PRWS(:,:,:) = PRWS(:,:,:) - DYF( MZM( MYM(PRHODJ) * PINV_PDYY) * ZFLX)
  END IF
END IF
#else
IF (.NOT. L2D) THEN 
  IF (.NOT. LFLAT) THEN
    CALL MYM_DEVICE(PRHODJ, ZTMP1_DEVICE)
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * PINV_PDYY(JI,JJ,JK)
    END DO !CONCURRENT  
    !$acc end kernels
    CALL MZM_DEVICE(ZTMP2_DEVICE, ZTMP1_DEVICE)
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP1_DEVICE(JI,JJ,JK) * ZFLX(JI,JJ,JK)
    END DO !CONCURRENT   
    !$acc end kernels
    CALL DYF_DEVICE( ZTMP2_DEVICE, ZTMP1_DEVICE )
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK)  = ZFLX(JI,JJ,JK) *PDZY(JI,JJ,JK) 
    END DO !CONCURRENT   
    !$acc end kernels
    CALL MZF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZTMP3_DEVICE(JI,JJ,JK) * PINV_PDYY(JI,JJ,JK)
    END DO !CONCURRENT  
    !$acc end kernels
    CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP3_DEVICE)
    CALL MZF_DEVICE( PDZZ, ZTMP2_DEVICE )
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP4_DEVICE(JI,JJ,JK) = PRHODJ(JI,JJ,JK) * ZTMP3_DEVICE(JI,JJ,JK) / ZTMP2_DEVICE(JI,JJ,JK)
    END DO !CONCURRENT   
    !$acc end kernels
    CALL DZM_DEVICE( ZTMP4_DEVICE, ZTMP2_DEVICE )
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       PRWS(JI,JJ,JK) = PRWS(JI,JJ,JK)  &
            - ZTMP1_DEVICE(JI,JJ,JK)          &
            + ZTMP2_DEVICE(JI,JJ,JK)
    END DO !CONCURRENT
    !$acc end kernels
  ELSE
    CALL MYM_DEVICE(PRHODJ, ZTMP1_DEVICE)
    !$acc kernels
    ZTMP2_DEVICE = ZTMP1_DEVICE * PINV_PDYY
    !$acc end kernels
    CALL MZM_DEVICE(ZTMP2_DEVICE, ZTMP1_DEVICE)
    !$acc kernels
    ZTMP2_DEVICE = ZTMP1_DEVICE * ZFLX
    !$acc end kernels
    CALL DYF_DEVICE( ZTMP2_DEVICE, ZTMP1_DEVICE )
    !$acc kernels
    PRWS(:,:,:) = PRWS(:,:,:) - ZTMP1_DEVICE
    !$acc end kernels
  END IF
END IF
#endif
!
IF (KSPLT==1) THEN
  ! 
  !Contribution to the dynamic production of TKE:
  !
  IF (.NOT. L2D) THEN
#ifndef MNH_OPENACC
    ZWORK(:,:,:) =-MZF( MYF( ZFLX *( GZ_V_VW(PVM,PDZZ) + GY_W_VW_PWM ) ) )
#else
    CALL GZ_V_VW_DEVICE(1,IKU,1,PVM,PDZZ,ZTMP1_DEVICE)
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZTMP2_DEVICE(JI,JJ,JK) = ZFLX(JI,JJ,JK) *( ZTMP1_DEVICE(JI,JJ,JK) + GY_W_VW_PWM(JI,JJ,JK) )
    END DO !CONCURRENT   
    !$acc end kernels
    CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
    CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
    !$acc kernels
    !$acc loop independent collapse(3)
    DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU)
       ZWORK(JI,JJ,JK) = -ZTMP2_DEVICE(JI,JJ,JK)
    END DO !CONCURRENT   
    !$acc end kernels
#endif
  !
  !
  ! evaluate the dynamic production at w(IKB+1) in PDP(IKB)
#ifndef MNH_OPENACC
    ZWORK(:,:,IKB:IKB) = - MYF (                                               &
       ZFLX(:,:,IKB+1:IKB+1) *                                                 &
     (   (PVM(:,:,IKB+1:IKB+1)-PVM(:,:,IKB:IKB)) / MYM(PDZZ(:,:,IKB+1:IKB+1))  &
       + ( DYM( PWM(:,:,IKB+1:IKB+1) )                                         &
          -MYM(  (PWM(:,:,IKB+2:IKB+2)-PWM(:,:,IKB+1:IKB+1))                   &
                  /(PDZZ(:,:,IKB+2:IKB+2)+PDZZ(:,:,IKB+1:IKB+1))               &
                +(PWM(:,:,IKB+1:IKB+1)-PWM(:,:,IKB  :IKB  ))                   &
                  /(PDZZ(:,:,IKB+1:IKB+1)+PDZZ(:,:,IKB  :IKB  ))               &
              ) * PDZY(:,:,IKB+1:IKB+1)                                        &
         ) / (0.5*(PDYY(:,:,IKB+1:IKB+1)+PDYY(:,:,IKB:IKB)))                   &
     )                        )
#else
    CALL MYM_DEVICE( PDZZ(:,:,IKB+1:IKB+1), ZTMP1_DEVICE(:,:,1:1) )
    CALL DYM_DEVICE( PWM(:,:,IKB+1:IKB+1), ZTMP2_DEVICE(:,:,1:1) )
    !$acc kernels
    ZTMP3_DEVICE(:,:,1) = (PWM(:,:,IKB+2)-PWM(:,:,IKB+1)) &
		    /(PDZZ(:,:,IKB+2)+PDZZ(:,:,IKB+1))    &
		  +(PWM(:,:,IKB+1)-PWM(:,:,IKB))          &
		    /(PDZZ(:,:,IKB+1)+PDZZ(:,:,IKB))
    !$acc end kernels
    CALL MYM_DEVICE( ZTMP3_DEVICE(:,:,1:1), ZTMP4_DEVICE(:,:,1:1) )
    !$acc kernels
    ZTMP3_DEVICE(:,:,1) = ZFLX(:,:,IKB+1) *                    &
      (   (PVM(:,:,IKB+1)-PVM(:,:,IKB)) / ZTMP1_DEVICE(:,:,1)  &
	+ ( ZTMP2_DEVICE(:,:,1)                                &
	    - ZTMP4_DEVICE(:,:,1) * PDZY(:,:,IKB+1)            &
	  ) / (0.5*(PDYY(:,:,IKB+1)+PDYY(:,:,IKB)))            &
      )
    !$acc end kernels
    CALL MYF_DEVICE( ZTMP3_DEVICE(:,:,1:1), ZTMP1_DEVICE(:,:,1:1) )
    !$acc kernels
    ZWORK(:,:,IKB) = - ZTMP1_DEVICE(:,:,1)
    !$acc end kernels
#endif
  ENDIF
  !
  ! dynamic production computation
  IF (.NOT. L2D) THEN
    !$acc kernels
    PDP(:,:,:) = PDP(:,:,:) + ZWORK(:,:,:)  
    !$acc end kernels
  ENDIF
  !
END IF
!
! Storage in the LES configuration (addition to TURB_VER computation)
!
IF (LLES_CALL .AND. KSPLT==1) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef MNH_OPENACC
  CALL LES_MEAN_SUBGRID( MZF(MYF(ZFLX)), X_LES_SUBGRID_WV , .TRUE. )
  CALL LES_MEAN_SUBGRID( MZF(MYF(GZ_V_VW(PVM,PDZZ)*ZFLX)), X_LES_RES_ddxa_V_SBG_UaV , .TRUE.)
  CALL LES_MEAN_SUBGRID( MZF(MYF(GY_W_VW(PWM,PDYY,PDZZ,PDZY)*ZFLX)), X_LES_RES_ddxa_W_SBG_UaW , .TRUE.)
  CALL LES_MEAN_SUBGRID( MXF(GY_M_V(1,IKU,1,PTHLM,PDYY,PDZZ,PDZY)*MZF(ZFLX)), &
                         X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE.)
  IF (KRR>=1) THEN
    CALL LES_MEAN_SUBGRID( MXF(GY_M_V(1,IKU,1,PRM(:,:,:,1),PDYY,PDZZ,PDZY)*MZF(ZFLX)), &
                           X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE.)
  END IF
  DO JSV=1,NSV
    CALL LES_MEAN_SUBGRID( MXF(GY_M_V(1,IKU,1,PSVM(:,:,:,JSV),PDYY,PDZZ,PDZY)*MZF(ZFLX)), &
                           X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV), .TRUE.)
  END DO
#else
!$acc data copy(X_LES_SUBGRID_WV,X_LES_RES_ddxa_V_SBG_UaV,X_LES_RES_ddxa_W_SBG_UaW, &
!$acc &         X_LES_RES_ddxa_Thl_SBG_UaW,X_LES_RES_ddxa_Sv_SBG_UaW)
  !
  CALL MYF_DEVICE(ZFLX,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_WV , .TRUE. )
  !
  CALL GZ_V_VW_DEVICE(1,IKU,1,PVM,PDZZ,ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE = ZTMP1_DEVICE*ZFLX
  !$acc end kernels
  CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_RES_ddxa_V_SBG_UaV , .TRUE.)
  !
  CALL GY_W_VW_DEVICE(1,IKU,1,PWM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZFLX(:,:,:)
  !$acc end kernels
  CALL MYF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZTMP1_DEVICE, ZTMP2_DEVICE )
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_RES_ddxa_W_SBG_UaW , .TRUE.)
  !
  CALL GY_M_V_DEVICE(1,IKU,1,PTHLM,PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
  CALL MZF_DEVICE( ZFLX, ZTMP2_DEVICE )
!PW: kernel removed to work around a PGI 19.10 bug
!$acc update self(ZTMP1_DEVICE,ZTMP2_DEVICE)
!   !$acc kernels
  ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
!   !$acc end kernels
!$acc update device(ZTMP3_DEVICE)
  CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE.)
  !
  IF (KRR>=1) THEN
    !
!$acc data copy(X_LES_RES_ddxa_Rt_SBG_UaW)
    !
    CALL GY_M_V_DEVICE(1,IKU,1,PRM(:,:,:,1),PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
    CALL MZF_DEVICE( ZFLX, ZTMP2_DEVICE )
!PW: kernel removed to work around a PGI 19.10 bug
!$acc update self(ZTMP1_DEVICE,ZTMP2_DEVICE)
!     !$acc kernels
    ZTMP3_DEVICE(:,:,:) = ZTMP1_DEVICE(:,:,:)*ZTMP2_DEVICE(:,:,:)
!     !$acc end kernels
!$acc update device(ZTMP3_DEVICE)
    CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE.)
    !
!$acc end data
    !
  END IF
  !
  CALL MZF_DEVICE( ZFLX, ZTMP2_DEVICE )
  DO JSV=1,NSV
    CALL GY_M_V_DEVICE(1,IKU,1,PSVM(:,:,:,JSV),PDYY,PDZZ,PDZY,ZTMP1_DEVICE)
    !$acc kernels
    ZTMP3_DEVICE = ZTMP1_DEVICE*ZTMP2_DEVICE
    !$acc end kernels
    CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV), .TRUE.)
  END DO
  !
!$acc end data
#endif
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF

if ( mppdb_initialized ) then
  !Check all inout arrays
  call Mppdb_check( prvs, "Turb_hor_vw end:prvs" )
  call Mppdb_check( prws, "Turb_hor_vw end:prws" )
  call Mppdb_check( pdp,  "Turb_hor_vw end:pdp"  )
end if

!$acc end data

#ifndef MNH_OPENACC
deallocate( ZFLX, ZWORK, GY_W_VW_PWM )
#else
CALL MNH_REL_ZT3D ( IZFLX, IZWORK, IGY_W_VW_PWM, &
                    IZTMP1_DEVICE, IZTMP2_DEVICE, IZTMP3_DEVICE, IZTMP4_DEVICE )
#endif

!$acc end data

END SUBROUTINE TURB_HOR_VW
